#include "iterator.h"

void iterator_free(Iterator tt) {
  mxFree(tt.dd);
  mxFree(tt.ss);
  mxFree(tt.index);
  mxFree(tt.ostep);
  mxFree(tt.istep);
  if ( tt.jstep != NULL ) mxFree(tt.jstep);
}

Shape iterator_shape(Iterator tt) {
  Shape oo;
  oo.nn = tt.nn;
  oo.dd = tt.dd;
  oo.ss = tt.ss;
  return oo;
}

Index iterator_index(Iterator tt) {
  Index oo;
  oo.nn = tt.nn;
  oo.dd = tt.dd;
  return oo;
}


void iterator_computestep(int *step, Shape ii, Shape aa) {
  int kk, ll; 
  int normalstep, resetstep;

  normalstep = 1;
  resetstep  = 0;
  ll = 0;
  for ( kk = 0 ; kk < aa.nn ; kk ++ ) {
    if ( (ll >= ii.nn) || aa.dd[kk] < ii.dd[ll] ) {
      step[kk]    = - resetstep;
    } else {
      step[kk]    = normalstep - resetstep;
      resetstep  += (ii.ss[ll]-1) * normalstep;
      normalstep *= ii.ss[ll];
      ll ++;
    }
  }
}

Iterator iterator_1arg(Shape oo, Shape ii) {
  Shape jj;
  Iterator tt;

  jj         = shape_copy(ii);
  tt.nn      = jj.nn;
  tt.dd      = jj.dd;
  tt.ss      = jj.ss;
  tt.numelem = shape_numelem(jj);

  tt.index = mxMalloc(sizeof(int) * tt.nn);
  tt.ostep = mxMalloc(sizeof(int) * tt.nn);
  tt.istep = mxMalloc(sizeof(int) * tt.nn);
  tt.jstep = NULL;

  iterator_computestep(tt.ostep,oo,jj);
  iterator_computestep(tt.istep,ii,jj);
  return tt;
}

Iterator iterator_2arg(Shape oo, Shape ii, Shape jj) {
  Shape kk;
  Iterator tt;

  kk         = shape_union(ii,jj);
  tt.nn      = kk.nn;
  tt.dd      = kk.dd;
  tt.ss      = kk.ss;
  tt.numelem = shape_numelem(kk);

  tt.index = mxMalloc(sizeof(int) * tt.nn);
  tt.ostep = mxMalloc(sizeof(int) * tt.nn);
  tt.istep = mxMalloc(sizeof(int) * tt.nn);
  tt.jstep = mxMalloc(sizeof(int) * tt.nn);

  iterator_computestep(tt.ostep,oo,kk);
  iterator_computestep(tt.istep,ii,kk);
  iterator_computestep(tt.jstep,jj,kk);
  return tt;
}

Iterator iterator_permute(Shape ii, int *dd, int *perm) {
  Iterator tt;
  int kk, *normalstep, *resetstep, step;

  tt.nn      = ii.nn;
  tt.dd      = mxMalloc(sizeof(int) * tt.nn);
  tt.ss      = mxMalloc(sizeof(int) * tt.nn);
  tt.numelem = shape_numelem(ii);
  tt.index   = mxMalloc(sizeof(int) * tt.nn);
  tt.ostep   = mxMalloc(sizeof(int) * tt.nn);
  tt.istep   = mxMalloc(sizeof(int) * tt.nn);
  tt.jstep   = NULL;

  normalstep = mxMalloc(sizeof(int) * ii.nn);
  resetstep  = mxMalloc(sizeof(int) * ii.nn);
  step       = 1;
  for ( kk = 0 ; kk < ii.nn ; kk ++ ) {
    tt.ostep[kk]      = 1;
    tt.dd[kk]         = dd[kk];
    tt.ss[kk]         = ii.ss[perm[kk]];
    normalstep[kk]    = step;
    resetstep[kk]     = (ii.ss[kk]-1) * step;
    step             *= ii.ss[kk];
  }

  step = 0;
  for ( kk = 0 ; kk < ii.nn ; kk ++ ) {
    tt.istep[kk] = normalstep[perm[kk]] - step;
    step        += resetstep[perm[kk]];
  }
  mxFree(normalstep);
  mxFree(resetstep);
  return tt;
}

