#include "table.h"

/* basic functions */

const char *table_mxfields[4] = {"dd","ss","zz","ee"};

Table table_new(Shape ii) {
  Table oo;
  Shape jj;
  jj         = shape_copy(ii);
  oo.nn      = jj.nn;
  oo.dd      = jj.dd;
  oo.ss      = jj.ss;
  oo.numelem = shape_numelem(jj);
  oo.zz      = mxMalloc(sizeof(double));
  *oo.zz     = 0.0;
  oo.ee      = mxMalloc(sizeof(double) * oo.numelem);
  return oo;
}

void table_free(Table ii) {
  mxFree(ii.dd);
  mxFree(ii.ss);
  mxFree(ii.zz);
  mxFree(ii.ee);
}

Table table_copy(Table ii) {
  double *oep, *endp, *tep;
  int kk;
  Table oo;
  oo.nn       = ii.nn;
  oo.dd       = mxMalloc(sizeof(int) * oo.nn);
  oo.ss       = mxMalloc(sizeof(int) * oo.nn);
  for ( kk = 0 ; kk < oo.nn ; kk ++ ) {
    oo.dd[kk] = ii.dd[kk];
    oo.ss[kk] = ii.ss[kk];
  }
  oo.numelem  = ii.numelem;
  oo.zz       = mxMalloc(sizeof(double));
  *oo.zz      = *ii.zz;
  oo.ee       = mxMalloc(sizeof(double) * oo.numelem);
  tep         = ii.ee;
  oep         = oo.ee;
  for ( endp = oep + oo.numelem ; oep < endp ; ) {
    * ( oep ++ ) =  * ( tep ++ );
  }
  return oo;
}

Table table_readshape(Shape ii, double *zz, double *ee) {
  Table oo;
  double *oep, *endp;
  oo = table_new(ii);
  *oo.zz = *zz;
  oep = oo.ee;
  for ( endp = oep + oo.numelem ; oep < endp ; ) {
    * ( oep ++ ) = * ( ee ++ );
  }
  return oo;
}

Shape table_writeshape(Table ii) {
  return shape_copy(table_shape(ii));
}

Index table_writeindex(Table ii) {
  return index_copy(table_index(ii));
}

Shape table_shape(Table ii) {
  Shape oo;
  oo.nn = ii.nn;
  oo.dd = ii.dd;
  oo.ss = ii.ss;
  return oo;
}

Index table_index(Table ii) {
  Index oo;
  oo.nn = ii.nn;
  oo.dd = ii.dd;
  return oo;
}

Table table_readmxarray(const mxArray *marray) {
  Shape ii;
  Table oo;
  double *tep, *endp, *mep;
  ii       = shape_readmxarray(marray);
  oo       = table_new(ii);
  if ( mxIsClass(marray,"table") ) {
    *oo.zz = *mxGetPr(mxGetField(marray,0,"zz"));
    mep    = mxGetPr(mxGetField(marray,0,"ee"));
  } else {
    *oo.zz = 0.0;
    mep    = mxGetPr(marray);
  }
  tep = oo.ee;
  for ( endp = tep + oo.numelem ; tep < endp ; ) {
    * ( tep ++ ) = * ( mep ++ );
  }
  shape_free(ii);
  return oo;
}

mxArray *table_writemxarray(Table ii) {
  mxArray *marray, *oo;
  marray = mxCreateStructMatrix(1,1,4,table_mxfields);
  table_rewritemxarray(marray,ii);
  mexCallMATLAB(1,&oo,1,&marray,"table");
  mxDestroyArray(marray);
  return oo;
}

void table_rewritemxarray(mxArray *marray, Table ii) {
  mxArray *me;
  double *tep, *endp;
  double *mep;

  shape_rewritemxarray(marray,table_shape(ii));
  if ( ( me = mxGetField(marray,0,"zz") ) != NULL ) mxDestroyArray(me);
  if ( ( me = mxGetField(marray,0,"ee") ) != NULL ) mxDestroyArray(me);
  mxSetField(marray,0,"zz",mxCreateScalarDouble(*ii.zz));
  if ( ii.nn == 0 ) {
    me  = mxCreateScalarDouble(*ii.zz);
  } else {
    me  = mxCreateNumericArray(ii.nn,ii.ss,mxDOUBLE_CLASS,mxREAL);
  }
  mep = mxGetPr(me);
  tep = ii.ee;
  for ( endp = tep + ii.numelem ; tep < endp ; ) {
    * ( mep ++ ) = * ( tep ++ ); 
  }
  mxSetField(marray,0,"ee",me);
}
/* unary operators */

void table_zeros(Table ii) {
  double *tep, *endp;
  *ii.zz = 0.0;
  for ( endp = ( tep = ii.ee ) + ii.numelem ; tep < endp ; ) {
    * ( tep ++ ) = 0.0;
  }
}

void table_ones(Table ii) {
  double *tep, *endp;
  *ii.zz = 0.0;
  for ( endp = ( tep = ii.ee ) + ii.numelem ; tep < endp ; ) {
    * ( tep ++ ) = 1.0;
  }
}

void table_inf(Table ii) {
  double *tep, *endp;
  double inf;
  inf = mxGetInf();
  *ii.zz = 0.0;
  for ( endp = ( tep = ii.ee ) + ii.numelem ; tep < endp ; ) {
    * ( tep ++ ) = inf;
  }
}

void table_neginf(Table ii) {
  double *tep, *endp;
  double inf;
  inf = mxGetInf();
  *ii.zz = 0.0;
  for ( endp = ( tep = ii.ee ) + ii.numelem ; tep < endp ; ) {
    * ( tep ++ ) = -inf;
  }
}

void table_uminus(Table ii) {
  double *tep, *endp;
  for ( endp = ( tep = ii.ee ) + ii.numelem ; tep < endp ; tep ++ ) {
    * tep = - * tep;
  }
}

void table_sumnorm(Table ii) {
  double mm, dd, *tep, *endp;
  mm = 0;
  for ( endp = ( tep = ii.ee) + ii.numelem ; tep < endp ; ) {
    dd  = * ( tep ++ );
    mm += dd < 0.0 ? - dd : dd;
  }
  if ( mm > 0 ) {
    for ( endp = ( tep = ii.ee) + ii.numelem ; tep < endp ; ) {
      * ( tep ++ ) /= mm;
    }
    *ii.zz += log(mm);
  }
}

void table_maxnorm(Table ii) {
  double mm, dd, *tep, *endp;
  mm = -mxGetInf();
  for ( endp = ( tep = ii.ee) + ii.numelem ; tep < endp ; ) {
    dd  = * ( tep ++ );
    if ( mm < dd )  mm = dd;
  }
  if ( mm > 0 ) {
    for ( endp = ( tep = ii.ee) + ii.numelem ; tep < endp ; ) {
      * ( tep ++ ) /= mm;
    }
    *ii.zz += log(mm);
  }
}

void table_maxabsnorm(Table ii) {
  double mm, dd, *tep, *endp;
  mm = -mxGetInf();
  for ( endp = ( tep = ii.ee) + ii.numelem ; tep < endp ; ) {
    dd = * ( tep ++ );
    if ( dd < 0.0 ) dd = - dd;
    if ( mm < dd )  mm = dd;
  }
  if ( mm > 0 ) {
    for ( endp = ( tep = ii.ee) + ii.numelem ; tep < endp ; ) {
      * ( tep ++ ) /= mm;
    }
    *ii.zz += log(mm);
  }
}


/* 2 argument elementwise operators */

/* assumes existence of table_basic_##elemop_name */
#define table_op_2arg(elemop_name)                                           \
Table table_##elemop_name(Table ii, Table jj) {                              \
  Table oo;                                                                  \
  Shape si, sj, so;                                                          \
  Iterator tt;                                                               \
  si = table_shape(ii);                                                      \
  sj = table_shape(jj);                                                      \
  so = shape_union(si,sj);                                                   \
  oo = table_new(so);                                                        \
  tt = iterator_2arg(so,si,sj);                                              \
    table_basic_##elemop_name(oo,ii,jj,tt);                                  \
  shape_free(so);                                                            \
  iterator_free(tt);                                                         \
  return oo;                                                                 \
}

/* assumes existence of table_basic_##elemop_name##_elemop */
#define table_basic_op_2arg_primitive(elemop_name,modii,modjj)               \
  if ( tt.nn == 0 ) {                                                        \
    table_##elemop_name##_elemop(*oo.ee, *ii.ee modii, *jj.ee modjj);        \
    return;                                                                  \
  }                                                                          \
                                                                             \
  for ( kk = 0 ; kk < tt.nn ; kk ++ )                                        \
    tt.index[kk] = 0;                                                        \
                                                                             \
  size0  = *tt.ss;                                                           \
  index0 = 0;                                                                \
                                                                             \
  oep = oo.ee; ostep0 = *tt.ostep;                                           \
  iep = ii.ee; istep0 = *tt.istep;                                           \
  jep = jj.ee; jstep0 = *tt.jstep;                                           \
                                                                             \
  while (1) {                                                                \
    table_##elemop_name##_elemop(*oep, *iep modii, *jep modjj);              \
                                                                             \
    if ( (++index0) < size0 ) {                                              \
      oep += ostep0;                                                         \
      iep += istep0;                                                         \
      jep += jstep0;                                                         \
    } else {                                                                 \
      index0 = 0;                                                            \
      for ( kk = 1 ; kk < tt.nn ; kk ++ ) {                                  \
        if ( (++ tt.index[kk]) < tt.ss[kk] ) {                               \
          oep += tt.ostep[kk];                                               \
          iep += tt.istep[kk];                                               \
          jep += tt.jstep[kk];                                               \
          break;                                                             \
        } else {                                                             \
          tt.index[kk] = 0;                                                  \
        }                                                                    \
      }                                                                      \
      if ( kk >= tt.nn )                                                     \
        break;                                                               \
    }                                                                        \
  } 

/* Types of 2arg elementwise operators: comparative, additive, 
 * multiplicative. Assumes table_##elemop_name##_elemop */
#define table_basic_op_comparative2arg(elemop_name)                          \
void table_basic_##elemop_name(Table oo, Table ii, Table jj, Iterator tt) {  \
  register double *oep, *iep, *jep;                                          \
  register int size0, index0, ostep0, istep0, jstep0;                        \
  register int kk;                                                           \
  register double power;                                                     \
                                                                             \
  if ( tt.numelem == 0 ) return;                                \
  *oo.zz = 0.0;                                                              \
  if ( *ii.zz == *jj.zz ) {                                                  \
    table_basic_op_2arg_primitive(elemop_name, , );                          \
  } else if ( *ii.zz < *jj.zz ) {                                            \
    power = exp(*ii.zz-*jj.zz);                                              \
    table_basic_op_2arg_primitive(elemop_name, *power, );                    \
  } else {                                                                   \
    power = exp(*jj.zz-*ii.zz);                                              \
    table_basic_op_2arg_primitive(elemop_name, , *power);                    \
  }                                                                          \
}
          

#define table_basic_op_additive2arg(elemop_name)                             \
void table_basic_##elemop_name(Table oo, Table ii, Table jj, Iterator tt) {  \
  register double *oep, *iep, *jep;                                          \
  register int size0, index0, ostep0, istep0, jstep0;                        \
  register int kk;                                                           \
  register double power;                                                     \
                                                                             \
  if ( tt.numelem == 0 ) return;                                             \
  if ( *ii.zz == *jj.zz ) {                                                  \
    *oo.zz = *jj.zz;                                                         \
    table_basic_op_2arg_primitive(elemop_name, , );                          \
  } else if ( *ii.zz < *jj.zz ) {                                            \
    *oo.zz = *jj.zz;                                                         \
    power = exp(*ii.zz-*jj.zz);                                              \
    table_basic_op_2arg_primitive(elemop_name, *power, );                    \
  } else {                                                                   \
    *oo.zz = *ii.zz;                                                         \
    power = exp(*jj.zz-*ii.zz);                                              \
    table_basic_op_2arg_primitive(elemop_name, , *power);                    \
  }                                                                          \
}                                                                           

/* assumes existence of table_basic_##elemop_name##_zzop as well */
#define table_basic_op_multiplicative2arg(elemop_name)                       \
void table_basic_##elemop_name(Table oo, Table ii, Table jj, Iterator tt) {  \
  register double *oep, *iep, *jep;                                          \
  register int size0, index0, ostep0, istep0, jstep0;                        \
  register int kk;                                                           \
  register double power;                                                     \
                                                                             \
  if ( tt.numelem == 0 ) return;                                             \
  table_##elemop_name##_zzop(*oo.zz,*ii.zz,*jj.zz);                          \
  table_basic_op_2arg_primitive(elemop_name, , );                            \
} 

/* comparative 2 argument elementwise operators */

#define table_eq_elemop(oo,ii,jj) (oo) = ((ii) == (jj))
table_basic_op_comparative2arg(eq)
table_op_2arg(eq)

#define table_ne_elemop(oo,ii,jj) (oo) = ((ii) != (jj))
table_basic_op_comparative2arg(ne)
table_op_2arg(ne)

#define table_lt_elemop(oo,ii,jj) (oo) = ((ii) < (jj))
table_basic_op_comparative2arg(lt)
table_op_2arg(lt)

#define table_gt_elemop(oo,ii,jj) (oo) = ((ii) > (jj))
table_basic_op_comparative2arg(gt)
table_op_2arg(gt)

#define table_le_elemop(oo,ii,jj) (oo) = ((ii) <= (jj))
table_basic_op_comparative2arg(le)
table_op_2arg(le)

#define table_ge_elemop(oo,ii,jj) (oo) = ((ii) >= (jj))
table_basic_op_comparative2arg(ge)
table_op_2arg(ge)

/* additive 2 argument elementwise operators */

#define table_plus_elemop(oo,ii,jj) (oo) = (ii) + (jj)
table_basic_op_additive2arg(plus)
table_op_2arg(plus)

#define table_minus_elemop(oo,ii,jj) (oo) = (ii) - (jj)
table_basic_op_additive2arg(minus)
table_op_2arg(minus)

#define table_max_elemop(oo,ii,jj) (oo) = (ii)<(jj) ? (jj) : (ii) 
table_basic_op_additive2arg(max)
table_op_2arg(max)

#define table_min_elemop(oo,ii,jj) (oo) = (ii)<(jj) ? (ii) : (jj) 
table_basic_op_additive2arg(min)
table_op_2arg(min)

/* multiplicative 2 argument elementwise operators */

#define table_times_zzop(oo,ii,jj) (oo) = (ii) + (jj)
#define table_times_elemop(oo,ii,jj) (oo) = (ii) * (jj)
table_basic_op_multiplicative2arg(times)
table_op_2arg(times)

#define table_ldivide_zzop(oo,ii,jj) (oo) = (jj) - (ii)
#define table_ldivide_elemop(oo,ii,jj) \
  (oo) = ( (ii)==0.0 && (jj)==0.0 ) ? 0.0 : (jj) / (ii)
table_basic_op_multiplicative2arg(ldivide)
table_op_2arg(ldivide)

#define table_rdivide_zzop(oo,ii,jj) (oo) = (ii) - (jj)
#define table_rdivide_elemop(oo,ii,jj) \
  (oo) = ( (ii)==0.0 && (jj)==0.0 ) ? 0.0 : (ii) / (jj)
table_basic_op_multiplicative2arg(rdivide)
table_op_2arg(rdivide)


/* exponentiative 2 argument elementwise operators */

void table_basic_power(Table oo, Table ii, Table jj, Iterator tt) {
  register double *oep, *iep, *jep;                                          
  register int size0, index0, ostep0, istep0, jstep0;                        
  register int kk;                                                           
  register double poweri, powerj;
  Table i2, j2;                                                            

  if ( tt.numelem == 0 ) return;                                

  i2 = table_copy(ii); table_maxnorm(i2);     
  j2 = table_copy(jj); table_maxnorm(j2);      
                                                                             
  poweri = *i2.zz;
  powerj = exp(*j2.zz);
  *oo.zz = poweri * powerj;
                                                                             
  if ( tt.nn == 0 ) {                                                 
    *oo.ee = 1.0;
    table_free(i2);
    table_free(j2);
    return;                                                                  
  }                                                                          

  for ( kk = 0 ; kk < tt.nn ; kk ++ )                                 
    tt.index[kk] = 0;                                                        
                                                                             
  size0  = *tt.ss;                                                   
  index0 = 0;                                                                
                                                                             
  oep = oo.ee; ostep0 = *tt.ostep;                                           
  iep = i2.ee; istep0 = *tt.istep;                                           
  jep = j2.ee; jstep0 = *tt.jstep;                                           
                                                                             
  while (1) {                                                                
    *oep = pow(*iep, (*jep * powerj)) * exp(poweri * ( *jep - 1.0 ) * powerj);
                                                                             
    if ( (++index0) < size0 ) {                                              
      oep += ostep0;                                                         
      iep += istep0;                                                         
      jep += jstep0;                                                         
    } else {                                                                 
      index0 = 0;                                                            
      for ( kk = 1 ; kk < tt.nn ; kk ++ ) {                           
        if ( (++ tt.index[kk]) < tt.ss[kk] ) {                       
          oep += tt.ostep[kk];                                               
          iep += tt.istep[kk];                                               
          jep += tt.jstep[kk];                                               
          break;                                                             
        } else {                                                             
          tt.index[kk] = 0;                                                  
        }                                                                    
      }                                                                      
      if ( kk >= tt.nn )                                              
        break;                                                               
    }                                                                        
  } 
  table_free(i2);
  table_free(j2);
}

table_op_2arg(power)

/* marginalization operators */

/* assumes table_##elemop_name##_ooinit and table_##elemop_name */
#define table_op_marg(elemop_name)                                           \
Table table_##elemop_name(Table ii, Index jj) {                             \
  Table oo;                                                                  \
  Shape si, so;                                                              \
  Iterator tt;                                                               \
  si = table_shape(ii);                                                      \
  so = shape_indexdiff(si,jj);                                              \
  oo = table_new(so);                                                        \
  tt = iterator_1arg(so,si);                                                 \
    table_##elemop_name##_ooinit(oo);                                        \
    table_basic_##elemop_name(oo,ii,tt);                                     \
  shape_free(so);                                                            \
  iterator_free(tt);                                                         \
  return oo;                                                                 \
}                                                                            \
                                                                             \
Table table_##elemop_name##to(Table ii, Index jj) {                         \
  Table oo;                                                                  \
  Shape si, so;                                                              \
  Iterator tt;                                                               \
  si = table_shape(ii);                                                      \
  so = shape_ref(si,jj);                                                     \
  oo = table_new(so);                                                        \
  tt = iterator_1arg(so,si);                                                 \
    table_##elemop_name##_ooinit(oo);                                        \
    table_basic_##elemop_name(oo,ii,tt);                                     \
  shape_free(so);                                                            \
  iterator_free(tt);                                                         \
  return oo;                                                                 \
}

/* assumes table_##elemop_name##_elemop */
#define table_basic_op_marg_primitive(elemop_name,modi)                      \
                                                                             \
  if ( tt.nn == 0 ) {                                                        \
    table_##elemop_name##_elemop(*oo.ee,*ii.ee,modi);                        \
    return;                                                                  \
  }                                                                          \
                                                                             \
  for ( kk = 0 ; kk < tt.nn ; kk ++ )                                        \
    tt.index[kk] = 0;                                                        \
                                                                             \
  size0  = *tt.ss;                                                           \
  index0 = 0;                                                                \
                                                                             \
  oep = oo.ee; ostep0 = *tt.ostep;                                           \
  iep = ii.ee; istep0 = *tt.istep;                                           \
                                                                             \
  while (1) {                                                                \
    table_##elemop_name##_elemop(*oep,*iep,modi);                            \
                                                                             \
    if ( (++index0) < size0 ) {                                              \
      oep += ostep0;                                                         \
      iep += istep0;                                                         \
    } else {                                                                 \
      index0 = 0;                                                            \
      for ( kk = 1 ; kk < tt.nn ; kk ++ ) {                                  \
        if ( (++ tt.index[kk]) < tt.ss[kk] ) {                               \
          oep += tt.ostep[kk];                                               \
          iep += tt.istep[kk];                                               \
          break;                                                             \
        } else {                                                             \
          tt.index[kk] = 0;                                                  \
        }                                                                    \
      }                                                                      \
      if ( kk >= tt.nn )                                                     \
        break;                                                               \
    }                                                                        \
  } 

/* assumes table_##elemop_name##_elemop */
#define table_basic_op_additivemarg(elemop_name)                             \
void table_basic_##elemop_name(Table oo, Table ii, Iterator tt) {            \
  register double *oep, *iep;                                                \
  register int size0, index0, ostep0, istep0;                                \
  register int kk;                                                           \
  register double tmp, power;                                                \
                                                                             \
  if ( tt.numelem == 0 ) return;                                             \
  if ( *oo.zz == *ii.zz ) {                                                  \
    table_basic_op_marg_primitive(elemop_name, );                            \
  } else if ( *oo.zz < *ii.zz ) {                                            \
    power = exp(*oo.zz-*ii.zz);                                              \
    *oo.zz = *ii.zz;                                                         \
    for ( iep = ( oep = oo.ee ) + oo.numelem ; oep < iep ; ) {               \
      *(oep++) *= power;                                                     \
    }                                                                        \
    table_basic_op_marg_primitive(elemop_name, );                            \
  } else {                                                                   \
    power = exp(*ii.zz-*oo.zz);                                              \
    table_basic_op_marg_primitive(elemop_name, *power);                      \
  }                                                                          \
}                                                                            

#define table_basic_op_multiplicativemarg(elemop_name)                       \
void table_basic_##elemop_name(Table oo, Table ii, Iterator tt) {            \
  register double *oep, *iep;                                                \
  register int size0, index0, ostep0, istep0;                                \
  register int kk;                                                           \
  register double tmp, power;                                                \
                                                                             \
  if ( tt.numelem == 0 ) return;                                             \
  *oo.zz += *ii.zz * ii.numelem / oo.numelem;                                \
  table_basic_op_marg_primitive(elemop_name, );                              \
}                                                                            

/* additive marginalization (contractive) operators */

#define table_sum_ooinit table_zeros
#define table_sum_elemop(oo,ii,modi) (oo) += (ii) modi
table_basic_op_additivemarg(sum)
table_op_marg(sum)

#define table_maxmarg_ooinit table_neginf
#define table_maxmarg_elemop(oo,ii,modi) if ((oo) < (tmp=(ii)modi)) (oo) = tmp
table_basic_op_additivemarg(maxmarg)
table_op_marg(maxmarg)

#define table_minmarg_ooinit table_inf
#define table_minmarg_elemop(oo,ii,modi) if ((oo) > (tmp=(ii)modi)) (oo) = tmp 
table_basic_op_additivemarg(minmarg)
table_op_marg(minmarg)

/* multiplicative marginalization (contractive) operators */

#define table_prod_ooinit table_ones
#define table_prod_elemop(oo,ii,modi) (oo) *= (ii) modi
table_basic_op_multiplicativemarg(prod)
table_op_marg(prod)


/* tensor operators */

/* assumes table_##elemop_name##_ooinit and table_##elemop_name */
#define table_op_tensor(elemop_name)                                         \
Table table_##elemop_name(Table ii, Table jj, Index kk) {                   \
  Table oo;                                                                  \
  Shape si, sj, su, so;                                                      \
  Iterator tt;                                                               \
  si = table_shape(ii);                                                      \
  sj = table_shape(jj);                                                      \
  su = shape_union(si,sj);                                                   \
  so = shape_indexdiff(su,kk);                                              \
  oo = table_new(so);                                                        \
  table_##elemop_name##_ooinit(oo);                                          \
  tt = iterator_2arg(so,si,sj);                                              \
    table_basic_##elemop_name(oo,ii,jj,tt);                                  \
  shape_free(su);                                                            \
  shape_free(so);                                                            \
  iterator_free(tt);                                                         \
  return oo;                                                                 \
}

#define table_basic_op_tensor_primitive(elemop_name,modi,modj,modij)         \
  if ( tt.nn == 0 ) {                                                        \
    table_##elemop_name##_elemop(*oo.ee,*ii.ee,*jj.ee,modi,modj,modij);      \
    return;                                                                  \
  }                                                                          \
                                                                             \
  for ( kk = 0 ; kk < tt.nn ; kk ++ )                                        \
    tt.index[kk] = 0;                                                       \
                                                                             \
  size0  = *tt.ss;                                                           \
  index0 = 0;                                                               \
                                                                             \
  oep = oo.ee; ostep0 = *tt.ostep;                                           \
  iep = ii.ee; istep0 = *tt.istep;                                           \
  jep = jj.ee; jstep0 = *tt.jstep;                                           \
                                                                             \
  while (1) {                                                                \
    table_##elemop_name##_elemop(*oep,*iep,*jep,modi,modj,modij);            \
                                                                             \
    if ( (++index0) < size0 ) {                                             \
      oep += ostep0;                                                         \
      iep += istep0;                                                         \
      jep += jstep0;                                                         \
    } else {                                                                 \
      index0 = 0;                                                           \
      for ( kk = 1 ; kk < tt.nn ; kk ++ ) {                                  \
        if ( (++ tt.index[kk]) < tt.ss[kk] ) {                              \
          oep += tt.ostep[kk];                                               \
          iep += tt.istep[kk];                                               \
          jep += tt.jstep[kk];                                               \
          break;                                                             \
        } else {                                                             \
          tt.index[kk] = 0;                                                 \
        }                                                                    \
      }                                                                      \
      if ( kk >= tt.nn )                                                     \
        break;                                                               \
    }                                                                        \
  } 

/* Types of tensor operators: addadd, addmult.
 * Assumes table_##elemop_name##_elemop. */
#define table_basic_op_addaddtensor(elemop_name)                             \
void table_basic_##elemop_name(Table oo, Table ii, Table jj, Iterator tt) {  \
  register double *oep, *iep, *jep;                                          \
  register int size0, index0, ostep0, istep0, jstep0;                        \
  register int kk;                                                           \
  register double tmp, power1, power2;                                       \
                                                                             \
  if ( tt.numelem == 0 ) return;                                             \
  if ( *ii.zz == *jj.zz ) {                                                  \
    if ( *oo.zz == *ii.zz ) {                                                \
      table_basic_op_tensor_primitive(elemop_name, , , );                    \
    } else if ( *oo.zz < *ii.zz ) {                                          \
      power1 = exp(*oo.zz-*ii.zz);                                           \
      *oo.zz = *ii.zz;                                                       \
      for ( iep = ( oep = oo.ee ) + oo.numelem ; oep < iep ; ) {             \
        *(oep++) *= power1;                                                  \
      }                                                                      \
      table_basic_op_tensor_primitive(elemop_name, , , );                    \
    } else {                                                                 \
      power1 = exp(*ii.zz-*oo.zz);                                           \
      table_basic_op_tensor_primitive(elemop_name, , , *power1);             \
    }                                                                        \
  } else if ( *ii.zz < *jj.zz ) {                                            \
    power1 = exp(*ii.zz-*jj.zz);                                             \
    if ( *oo.zz == *jj.zz ) {                                                \
      table_basic_op_tensor_primitive(elemop_name, *power1, , );             \
    } else if ( *oo.zz < *jj.zz ) {                                          \
      power2 = exp(*oo.zz-*jj.zz);                                           \
      *oo.zz = *jj.zz;                                                       \
      for ( iep = ( oep = oo.ee ) + oo.numelem ; oep < iep ; ) {             \
        *(oep++) *= power2;                                                  \
      }                                                                      \
      table_basic_op_tensor_primitive(elemop_name, *power1, , );             \
    } else {                                                                 \
      power2 = exp(*jj.zz-*oo.zz);                                           \
      table_basic_op_tensor_primitive(elemop_name, *power1, , *power2);      \
    }                                                                        \
  } else {                                                                   \
    power1 = exp(*jj.zz-*ii.zz);                                             \
    if ( *oo.zz == *ii.zz ) {                                                \
      table_basic_op_tensor_primitive(elemop_name, , *power1, );             \
    } else if ( *oo.zz < *ii.zz ) {                                          \
      power2 = exp(*oo.zz-*ii.zz);                                           \
      *oo.zz = *ii.zz;                                                       \
      for ( iep = ( oep = oo.ee ) + oo.numelem ; oep < iep ; ) {             \
        *(oep++) *= power2;                                                  \
      }                                                                      \
      table_basic_op_tensor_primitive(elemop_name, , *power1, );             \
    } else {                                                                 \
      power2 = exp(*ii.zz-*oo.zz);                                           \
      table_basic_op_tensor_primitive(elemop_name, , *power1, *power2);      \
    }                                                                        \
  }                                                                          \
}                                                                            

#define table_basic_op_addmulttensor(elemop_name)                            \
void table_basic_##elemop_name(Table oo, Table ii, Table jj, Iterator tt) {  \
  register double *oep, *iep, *jep;                                          \
  register int size0, index0, ostep0, istep0, jstep0;                       \
  register int kk;                                                           \
  register double tmp, power1, power2;                                       \
                                                                             \
  if ( tt.numelem == 0 ) return;                                             \
  power1 = *ii.zz + *jj.zz;                                                  \
  if ( *oo.zz == power1 ) {                                                  \
    table_basic_op_tensor_primitive(elemop_name, , , );                      \
  } else if ( *oo.zz < power1 ) {                                            \
    power2 = exp(*oo.zz-power1);                                             \
    *oo.zz = power1;                                                         \
    for ( iep = ( oep = oo.ee ) + oo.numelem ; oep < iep ; ) {               \
      *(oep++) *= power2;                                                    \
    }                                                                        \
    table_basic_op_tensor_primitive(elemop_name, , , );                      \
  } else {                                                                   \
    power2 = exp(power1-*oo.zz);                                             \
    table_basic_op_tensor_primitive(elemop_name, , , *power2);               \
  }                                                                          \
} 

/* additive-additive 2 argument tensor operators */

#define table_maxsum_ooinit table_neginf
#define table_maxsum_elemop(oo,ii,jj,modi,modj,modij) \
  if ( (oo) < (tmp=((ii) modi)+((jj) modj) modij) ) (oo) = tmp
table_basic_op_addaddtensor(maxsum)
table_op_tensor(maxsum)

#define table_minsum_ooinit table_inf
#define table_minsum_elemop(oo,ii,jj,modi,modj,modij) \
  if ( (oo) > (tmp=((ii) modi)+((jj) modj) modij) ) (oo) = tmp
table_basic_op_addaddtensor(minsum)
table_op_tensor(minsum)

/* additive-multiplicative 2 argument tensor operators */

#define table_sumprod_ooinit table_zeros
#define table_sumprod_elemop(oo,ii,jj,modi,modj,modij) \
  (oo) += (((ii) modi) * ((jj) modj) modij)
table_basic_op_addmulttensor(sumprod)
table_op_tensor(sumprod)

#define table_maxprod_ooinit table_neginf
#define table_maxprod_elemop(oo,ii,jj,modi,modj,modij) \
  if ( (oo) < (tmp=((ii) modi)*((jj) modj) modij) ) (oo) = tmp
table_basic_op_addmulttensor(maxprod)
table_op_tensor(maxprod)

/* additive-multiplicative einstein-style 2 argument tensor operators */

Table table_mtimes(Table ii, Table jj) {
  Table oo;
  Shape so;
  Index kk;
  so = shape_intersect(table_shape(ii),table_shape(jj));
  kk = shape_index(so);
    oo = table_sumprod(ii,jj,kk);
  shape_free(so);
  return oo;
}

/* permuting dimensions */

void table_basic_permute(Table oo, Table ii, Iterator tt) {
  register double *oep, *iep;
  register int size0, index0, ostep0, istep0;
  register int kk;
  register double tmp, power;

  if ( tt.numelem == 0 || tt.nn == 0 ) return;

  *oo.zz      = *ii.zz;

  for ( kk = 0 ; kk < tt.nn ; kk ++ )
    tt.index[kk] = 0;
  size0  = *tt.ss;
  index0 = 0;

  oep = oo.ee; ostep0 = *tt.ostep;
  iep = ii.ee; istep0 = *tt.istep;

  while (1) {
    *oep = *iep;

    if ( (++index0) < size0 ) {
      oep += ostep0;
      iep += istep0;
    } else {
      index0 = 0;
      for ( kk = 1 ; kk < tt.nn ; kk ++ ) {
        if ( (++ tt.index[kk]) < tt.ss[kk] ) {
          oep += tt.ostep[kk];
          iep += tt.istep[kk];
          break;
        } else {
          tt.index[kk] = 0;
        }
      }
      if ( kk >= tt.nn )
        break;
    }
  }
}

Table table_permute(Table ii, int ndim, int *idim, int *odim) {
  Table oo;
  Iterator tt;
  int ki, ko, thisisin, mm;
  int *dim, *perm;

  dim  = mxMalloc(sizeof(int) * ii.nn);
  perm = mxMalloc(sizeof(int) * ii.nn);

  /* identify correspondences */
  for ( ki = 0 ; ki < ii.nn ; ki ++ ) {
    thisisin = 0;
    for ( ko = 0 ; ko < ndim ; ko ++ ) {
      if ( idim[ko] == ii.dd[ki] ) {
        dim[ki] = odim[ko];
        thisisin = 1;
        break;
      }
    }
    if (! thisisin) dim[ki] = ii.dd[ki];
    perm[ki] = ki;
  }

  /* sort */
  for ( ki = 0 ; ki < ii.nn ; ki ++ ) {
    mm = ki;
    for ( ko = ki+1 ; ko < ii.nn ; ko ++ ) {
      if ( dim[mm] == dim[ko] ) {
        mexErrMsgTxt("Dimensional collision!");
      } else if ( dim[mm] > dim[ko] ) {
        mm = ko;
      }
    }
    /*switch-a-roo*/
    if ( mm != ki ) {
      dim[ki]  = (dim[mm]  += (dim[ki]  -= dim[mm]))  - dim[ki]; 
      perm[ki] = (perm[mm] += (perm[ki] -= perm[mm])) - perm[ki];
    }
  }

  tt = iterator_permute(table_shape(ii),dim,perm);
  oo = table_new(iterator_shape(tt));
  table_basic_permute(oo,ii,tt);
  mxFree(dim);
  mxFree(perm);
  iterator_free(tt);
  return oo;
}

