#ifndef __TABLE__
#define __TABLE__
#include "math.h"
#include "index.h"
#include "shape.h"
#include "iterator.h"

typedef struct {
  int nn;
  int *dd; 
  int *ss;
  int numelem;
  double *zz;
  double *ee;
} Table;

Table table_new(Shape shape);
void table_free(Table table);
Table table_copy(Table table);

Table table_readshape(Shape ii, double *zz, double *ee);
Shape table_writeshape(Table ii);
Shape table_shape(Table ii);
Index table_writeindex(Table ii);
Index table_index(Table ii);

Table table_readmxarray(const mxArray *marray);
mxArray *table_writemxarray(Table ii);
void table_rewritemxarray(mxArray *marray, Table ii);

void table_zeros(Table table);
void table_ones(Table table);
void table_inf(Table table);
void table_neginf(Table table);
void table_uminus(Table table);
void table_sumnorm(Table table);
void table_maxnorm(Table table);
void table_maxabsnorm(Table table);

Table table_eq(Table ii, Table jj);
Table table_ne(Table ii, Table jj);
Table table_lt(Table ii, Table jj);
Table table_gt(Table ii, Table jj);
Table table_le(Table ii, Table jj);
Table table_ge(Table ii, Table jj);

Table table_plus   (Table ii, Table jj);
Table table_minus  (Table ii, Table jj);
Table table_max    (Table ii, Table jj);
Table table_min    (Table ii, Table jj);
Table table_times  (Table ii, Table jj);
Table table_ldivide(Table ii, Table jj);
Table table_rdivide(Table ii, Table jj);
Table table_power  (Table ii, Table jj);

Table table_sum      (Table ii, Index dd);
Table table_sumto    (Table ii, Index dd);
Table table_maxmarg  (Table ii, Index dd);
Table table_maxmargto(Table ii, Index dd);
Table table_minmarg  (Table ii, Index dd);
Table table_minmargto(Table ii, Index dd);
Table table_prod     (Table ii, Index dd);
Table table_prodto   (Table ii, Index dd);

Table table_maxsum (Table ii, Table jj, Index dd);
Table table_minsum (Table ii, Table jj, Index dd);
Table table_sumprod(Table ii, Table jj, Index dd);
Table table_maxprod(Table ii, Table jj, Index dd);

Table table_mtimes (Table ii, Table jj);

Table table_permute (Table ii, int ndim, int *idim, int *odim);

void table_basic_eq(Table oo, Table ii, Table jj, Iterator tt);
void table_basic_ne(Table oo, Table ii, Table jj, Iterator tt);
void table_basic_lt(Table oo, Table ii, Table jj, Iterator tt);
void table_basic_gt(Table oo, Table ii, Table jj, Iterator tt);
void table_basic_le(Table oo, Table ii, Table jj, Iterator tt);
void table_basic_ge(Table oo, Table ii, Table jj, Iterator tt);

void table_basic_plus   (Table oo, Table ii, Table jj, Iterator tt);
void table_basic_minus  (Table oo, Table ii, Table jj, Iterator tt);
void table_basic_max    (Table oo, Table ii, Table jj, Iterator tt);
void table_basic_min    (Table oo, Table ii, Table jj, Iterator tt);
void table_basic_times  (Table oo, Table ii, Table jj, Iterator tt);
void table_basic_ldivide(Table oo, Table ii, Table jj, Iterator tt);
void table_basic_rdivide(Table oo, Table ii, Table jj, Iterator tt);
void table_basic_power  (Table oo, Table ii, Table jj, Iterator tt);

void table_basic_sum    (Table oo, Table ii, Iterator tt);
void table_basic_maxmarg(Table oo, Table ii, Iterator tt);
void table_basic_minmarg(Table oo, Table ii, Iterator tt);
void table_basic_prod   (Table oo, Table ii, Iterator tt);

void table_basic_maxsum (Table oo, Table ii, Table jj, Iterator tt);
void table_basic_minsum (Table oo, Table ii, Table jj, Iterator tt);
void table_basic_sumprod(Table oo, Table ii, Table jj, Iterator tt);
void table_basic_maxprod(Table oo, Table ii, Table jj, Iterator tt);

void table_basic_mtimes(Table oo, Table ii, Table jj, Iterator tt);

void table_basic_permute(Table oo, Table ii, Iterator tt);

#endif
