#ifndef __TABLE_PERMUTE__
#define __TABLE_PERMUTE__
#include "table.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  Table ii, oo;
  Index jj;
  int ndim, *odim, *idim, kk;
  double *dd;
  if ( nrhs >= 2 ) {
    ii = table_readmxarray(prhs[0]);
    if ( nrhs == 2 ) {
      if ( mxGetN(prhs[1]) != 2 ) mexErrMsgTxt("2-column vectors expected.");
      ndim = mxGetM(prhs[1]);
      dd = mxGetPr(prhs[1]);
      idim = mxMalloc(sizeof(int) * ndim);
      odim = mxMalloc(sizeof(int) * ndim);
      for ( kk = 0 ; kk < ndim ; idim[ kk ++ ] = * ( dd ++ ) );
      for ( kk = 0 ; kk < ndim ; odim[ kk ++ ] = * ( dd ++ ) );
    } else if ( nrhs == 3 ) {
      if ( mxIsClass(prhs[1],"index") ) {
        jj   = index_readmxarray(prhs[1]);
        ndim = jj.nn;
        idim = mxMalloc(sizeof(int) * ndim);
        for ( kk = 0 ; kk < ndim ; kk ++ ) idim[kk] = jj.dd[kk];
        index_free(jj);
      } else {
        dd = mxGetPr(prhs[1]);
        ndim = mxGetNumberOfElements(prhs[1]);
        idim = mxMalloc(sizeof(int) * ndim);
        for ( kk = 0 ; kk < ndim ; idim[ kk ++ ] = * ( dd ++ ) );
      }

      odim = mxMalloc(sizeof(int) * ndim);
      if ( mxIsClass(prhs[1],"index") ) {
        jj   = index_readmxarray(prhs[2]);
        if ( ndim != jj.nn )
          mexErrMsgTxt("Number of entries in permutation pairs must match.");
        for ( kk = 0 ; kk < ndim ; kk ++ ) odim[kk] = jj.dd[kk];
        index_free(jj);
      } else {
        dd = mxGetPr(prhs[2]);
        if ( ndim != mxGetNumberOfElements(prhs[2]) )
          mexErrMsgTxt("Number of entries in permutation pairs must match.");
        for ( kk = 0 ; kk < ndim ; kk ++ ) odim[kk] = * ( dd ++ );
      }
    } else {
      mexErrMsgTxt("Two or Three arguments expected.");
    }
    oo = table_permute(ii,ndim,idim,odim);
  } else {
      mexErrMsgTxt("Two or Three arguments expected.");
  }
  plhs[0] = mxDuplicateArray(prhs[0]);
  table_rewritemxarray(plhs[0],oo);
  mxFree(idim);
  mxFree(odim);
  table_free(ii);
  table_free(oo);
}      

#endif
